KEY = '0'
GLOBE = '1'
WARNING_SIGN = '2'
SERVER = '3'
PINNED_NOTE = '4'
SPEECH_BUBBLE = '5'
SQUARES = '6'
HANDWRITTEN_NOTE = '7'
GLOBE_PLUG = '8'
BUSINESS_CARD = '9'
GREEN_AND_WHITE_THINGY = '10'
CAMERA = '11'
INFRARED = '12'
KEYS = '13'
POWER_PLUG = '14'
FLATBED_SCANNER = '15'
GLOBE_STAR = '16'
CD_ROM = '17'
MONITOR = '18'
ENVELOPE = '19'
GEAR = '20'
CHECKLIST = '21'
NOTEPAD = '22'
DESKTOP = '23'
POWER_FLASH = '24'
FOLDER_ENVELOPE = '25'
FLOPPY_DISK = '26'
SERVER_2 = '27'
GREEN_DOT = '28'
MONITOR_KEY = '29'
SHELL = '30'
PRINTER = '31'
DASHBOARD = '32'
CROATIA = '33'
WRENCH = '34'
PC_INTERNET = '35'
ZIP_FILE = '36'
PERCENT_SIGN = '37'
SAMBA_SHARE = '38'
CLOCK = '39'
SEARCH = '40'
USED_TAMPON = '41'
MEMORY_STICK = '42'
RECYCLE_BIN = '43'
POST_IT = '44'
RED_CROSS = '45'
INFO_SIGN = '46'
CARDBOARD = '47'
FOLDER = '48'
FOLDER_OPEN = '49'
FOLDER_CUBE = '50'
LOCK_OPEN = '51'
LOCK_CLOSED = '52'
GREEN_CHECKMARK = '53'
FEATHER_PEN = '54'
POLAROID_PICTURE = '55'
BOOK_OPENED = '56'
UI = '57'
MANAGER = '58'
HAMMER = '59'
HOUSE = '60'
STAR = '61'
PENGUIN = '62'
FEATHER = '63'
APPLE = '64'
WIKIPEDIA = '65'
DOLLAR_SIGN = '66'
CERTIFICATE = '67'
SMARTPHONE = '68'
